ADD = 'add'
AND = 'and'
ANEWARRAY = 'anewarray'
ARRLEN = 'arrlen'
ARRLOAD = 'arrload'
ARRLOAD_OBJ = 'arrload_obj'
ARRSTORE = 'arrstore'
ARRSTORE_OBJ = 'arrstore_obj'
CHECKCAST = 'checkcast'
CONST = 'const'
CONSTNULL = 'constnull'
CONVERT = 'convert'
DIV = 'div'
DUP = 'dup'
DUP2 = 'dup2'
DUP2X1 = 'dup2x1'
DUP2X2 = 'dup2x2'
DUPX1 = 'dupx1'
DUPX2 = 'dupx2'
FCMP = 'fcmp'
GETFIELD = 'getfield'
GETSTATIC = 'getstatic'
GOTO = 'goto'
IF_A = 'if_a'
IF_ACMP = 'if_acmp'
IF_I = 'if_i'
IF_ICMP = 'if_icmp'
IINC = 'iinc'
INSTANCEOF = 'instanceof'
INVOKEDYNAMIC = 'invokedynamic'
INVOKEINIT = 'invokeinit'
INVOKEINTERFACE = 'invokeinterface'
INVOKESPECIAL = 'invokespecial'
INVOKESTATIC = 'invokestatic'
INVOKEVIRTUAL = 'invokevirtual'
JSR = 'jsr'
LCMP = 'lcmp'
LDC = 'ldc'
LOAD = 'load'
MONENTER = 'monenter'
MONEXIT = 'monexit'
MUL = 'mul'
MULTINEWARRAY = 'multinewarray'
NEG = 'neg'
NEW = 'new'
NEWARRAY = 'newarray'
NOP = 'nop'
OR = 'or'
POP = 'pop'
POP2 = 'pop2'
PUTFIELD = 'putfield'
PUTSTATIC = 'putstatic'
REM = 'rem'
RET = 'ret'
RETURN = 'return'
SHL = 'shl'
SHR = 'shr'
STORE = 'store'
SUB = 'sub'
SWAP = 'swap'
SWITCH = 'switch'
THROW = 'throw'
TRUNCATE = 'truncate'
USHR = 'ushr'
XOR = 'xor'
